#ifndef DEFINE_H
#define DEFINE_H

#include <iostream>
#include <string>
#include <cassert>
#include <sstream>
#include <salt/vector.h>
#include <salt/gmath.h>
#include <soccer/soccertypes.h>


const std::string gServerPath   = "/sys/server/";
const std::string gScenePath    = "/usr/scene/";
//const std::string gSimScenePath = "/usr/simscene/";
const std::string gTeamName     = "apollo";

const unsigned int gTeamSize    = 6;
const static int JOINT_NUM      = 22;
const static int PART_NUM       = 23;

static const std::string ACTION_TURN            = "turn";
static const std::string ACTION_KICK            = "kick";
static const std::string ACTION_WALK_TO         = "walkto";
static const std::string ACTION_SIDE_WALK       = "sidewalk";
static const std::string ACTION_BLOCK           = "block";
static const std::string ACTION_GETUP           = "getup";
static const std::string ACTION                 = "action";


/*template<typename PTR> inline
void CheckPtr(PTR p, std::string msg)
{
    if (p.get() == 0)
    {
        std::cerr << "(CHECK_PTR) ERROR: " << msg
                  << " not found, interrupt..." << std::endl;
        exit(0);
    }
}*/

struct VisionSense
{
    /** distance perceptor to object */
    double distance;

    /** theta is the angle in the X-Y (horizontal) plane */
    double theta;

    /** phi is the latitude angle */
    double phi;

    salt::Vector3f localPos;

    salt::Vector3f realPos;

    /** init with some big number to find out bug easlier */
    VisionSense() : distance(-1000.0), theta(-1000.0), phi(-1000.0) {};
};

struct HingeJointSense
{
    /** joint angle */
    float angle;

    /** joint angle rate */
    float rate;

    HingeJointSense() : angle(0), rate(0) {};
};

struct UniversalJointSense
{
    /** joint angle axis 1*/
    float angle1;

    /** joint angle axis 2*/
    float angle2;

    /** joint angle rate axis 1*/
    float rate1;

    /** joint angle rate axis 2*/
    float rate2;

    UniversalJointSense() : angle1(0), angle2(0), rate1(0), rate2(0) {};
};

enum JointID
{
    JID_HEAD_2 = 1, //!! Head Joint 2 which belongs to head is on top of neck 
    JID_HEAD_1, 

    JID_LARM_1,
    JID_RARM_1,
    JID_LARM_2,
    JID_RARM_2,
    JID_LARM_3,
    JID_RARM_3,
    JID_LARM_4,
    JID_RARM_4,

    JID_LLEG_1,
    JID_RLEG_1,
    JID_LLEG_2,
    JID_RLEG_2,
    JID_LLEG_3,
    JID_RLEG_3,
    JID_LLEG_4,
    JID_RLEG_4,
    JID_LLEG_5,
    JID_RLEG_5,
    JID_LLEG_6,
    JID_RLEG_6 = 22
};

enum RobotPart
{
    PART_HEAD      = 1,
    PART_NECK, 
    PART_BODY,
    PART_LSHOULDER,
    PART_RSHOULDER,
    PART_LUPPERARM,
    PART_RUPPERARM,
    PART_LELBOW,
    PART_RELBOW,
    PART_LLOWERARM,
    PART_RLOWERARM,
    PART_LHIP_1,
    PART_RHIP_1,
    PART_LHIP_2,
    PART_RHIP_2,
    PART_LTHIGH,
    PART_RTHIGH,
    PART_LSHANK,
    PART_RSHANK,
    PART_LANKLE,
    PART_RANKLE,
    PART_LFOOT,
    PART_RFOOT      = 23
};

enum VisionObject
{
    FLAG_1_L    = 0,
    FLAG_1_R    = 1,
    FLAG_2_L    = 2,
    FLAG_2_R    = 3,
    GOAL_1_L    = 4,
    GOAL_1_R    = 5,
    GOAL_2_L    = 6,
    GOAL_2_R    = 7,
    BALL        = 8,

    TEAMMATE_1_HEAD  = 9,
    TEAMMATE_1_LLA  =10,
    TEAMMATE_1_RLA  =11,
    TEAMMATE_1_LF   =12,
    TEAMMATE_1_RF =13,

    
    TEAMMATE_2_HEAD =14 ,
    TEAMMATE_2_LLA	=15,
    TEAMMATE_2_RLA	=16,
    TEAMMATE_2_LF	=17,
    TEAMMATE_2_RF =18,

    TEAMMATE_3_HEAD =19 ,
    TEAMMATE_3_LLA	=20,
    TEAMMATE_3_RLA	=21,
    TEAMMATE_3_LF	=22,
    TEAMMATE_3_RF =23,

    TEAMMATE_4_HEAD =24 ,
    TEAMMATE_4_LLA	=25,
    TEAMMATE_4_RLA	=26,
    TEAMMATE_4_LF	=27,
    TEAMMATE_4_RF =28,

    TEAMMATE_5_HEAD =29 ,
    TEAMMATE_5_LLA	=30,
    TEAMMATE_5_RLA	=31,
    TEAMMATE_5_LF	=32,
    TEAMMATE_5_RF =33,

    TEAMMATE_6_HEAD =34 ,
    TEAMMATE_6_LLA	=35,
    TEAMMATE_6_RLA	=36,
    TEAMMATE_6_LF	=37,
    TEAMMATE_6_RF =38,
    
	OPPONENT_1_HEAD = 39,
	OPPONENT_1_LLA	=40,
	OPPONENT_1_RLA	=41,
	OPPONENT_1_LF	=42,
	OPPONENT_1_RF =43,

	
	OPPONENT_2_HEAD =44 ,
	OPPONENT_2_LLA	=45,
	OPPONENT_2_RLA	=46,
	OPPONENT_2_LF	=47,
	OPPONENT_2_RF =48,

	OPPONENT_3_HEAD =49 ,
	OPPONENT_3_LLA	=50,
	OPPONENT_3_RLA	=51,
	OPPONENT_3_LF	=52,
	OPPONENT_3_RF =53,

	OPPONENT_4_HEAD = 54,
	OPPONENT_4_LLA	=55,
	OPPONENT_4_RLA	=56,
	OPPONENT_4_LF	=57,
	OPPONENT_4_RF =58,

	OPPONENT_5_HEAD =59 ,
	OPPONENT_5_LLA	=60,
	OPPONENT_5_RLA	=61,
	OPPONENT_5_LF	=62,
	OPPONENT_5_RF =63,

	OPPONENT_6_HEAD = 64,
	OPPONENT_6_LLA	=65,
	OPPONENT_6_RLA	=66,
	OPPONENT_6_LF	=67,
	OPPONENT_6_RF =68,
};

struct GameState
{
    int unum;

    float match_time;

    TTeamIndex team_index;

    TPlayMode play_mode;

    float fieldlength;

    float fieldwidth;

    float goalwidth;

    float goaldepth;

    float goalheight;

    float penaltylength;

    float penaltywidth;

    GameState() : unum(0), 
                  match_time(0.0f), 
                  team_index(TI_NONE),
                  play_mode(PM_NONE),
                  fieldlength(-1.0f),
                  fieldwidth(-1.0f),
                  goalwidth(-1.0f),
                  goaldepth(-1.0f),
                  goalheight(-1.0f),
                  penaltylength(-1.0f),
                  penaltywidth(-1.0f)
    {
    }
};

struct  HearInfo
{
    float time;
    std::string sender;
    std::string msg;
};

#endif
