#include "kinematics.h"
#include <zeitgeist/logserver/logserver.h>
#include "geometry.h"

using namespace std;
using namespace salt;
using namespace zeitgeist;

Kinematics::Kinematics()
{
    FootLength          = 0.08;
    FootWidth           = 0.16;
    FootHeight          = 0.03;
    FootMass            = 0.2;

    ShankLength         = 0.08;
    ShankWidth          = 0.07;
    ShankHeight         = 0.11;
    ShankMass           = 0.225;

    ThighLength         = 0.07;
    ThighWidth          = 0.07;
    ThighHeight         = 0.14;
    ThighMass           = 0.275;

    TorsoLength         = 0.1;
    TorsoWidth          = 0.1;
    TorsoHeight         = 0.18;
    TorsoMass           = 1.2171;

    TorsoInitX          = 0.0f;
    TorsoInitY          = 0.0f;
    TorsoInitZ          = 0.385;


    Hip1RelTorso_X      = 0.055;
    Hip1RelTorso_Y      = -0.01;
    Hip1RelTorso_Z      = -0.115;

    Hip2RelHip1_X       = 0;
    Hip2RelHip1_Y       = 0;
    Hip2RelHip1_Z       = 0;

    ThighRelHip2_X      = 0;
    ThighRelHip2_Y      = 0.01;  //set 0.005 to check the tri-angle algorithm
    ThighRelHip2_Z      = -0.04;

    ShankRelThigh_X     = 0;
    ShankRelThigh_Y     = 0.005;
    ShankRelThigh_Z     = -0.125;

    AnkleRelShank_X     = 0;
    AnkleRelShank_Y     = -0.01;
    AnkleRelShank_Z     = -0.055;

    FootRelAnkle_X      = 0;
    FootRelAnkle_Y      = 0.03;
    FootRelAnkle_Z      = -0.035;

    Hip1Axis_X          = -0.7071;
    Hip1Axis_Y          = 0;
    Hip1Axis_Z          = 0.7071;

    SetupLinks();
}

Kinematics::~Kinematics()
{
}

void Kinematics::SetupLinks()
{
    uLINK[1].name       = "TORSO";
    uLINK[1].sister     = 0;
    uLINK[1].child      = 2;
    uLINK[1].mother     = 0;
    uLINK[1].a          = Vector3f(0, 0, 0);
    uLINK[1].c2a        = Vector3f(0, 0, 0);
    uLINK[1].b          = Vector3f(0, 0, 0);
    uLINK[1].q          = 0;
    uLINK[1].p          = Vector3f(TorsoInitX, TorsoInitY, TorsoInitZ);
    uLINK[1].R.Identity();


    uLINK[2].name       = "LHIP_1";
    uLINK[2].sister     = 8;
    uLINK[2].child      = 3;
    uLINK[2].mother     = 1;
    uLINK[2].a          = Vector3f(Hip1Axis_X, Hip1Axis_Y, -Hip1Axis_Z);
    uLINK[2].c2a        = Vector3f(0, 0, 0);
    uLINK[2].b          = Vector3f(-Hip1RelTorso_X, Hip1RelTorso_Y, Hip1RelTorso_Z) +
                          uLINK[2].c2a - uLINK[1].c2a;
    uLINK[2].q          = 0;

    uLINK[3].name       = "LHIP_2";
    uLINK[3].sister     = 0;
    uLINK[3].child      = 4;
    uLINK[3].mother     = 2;
    uLINK[3].a          = Vector3f(0, 1, 0); // Vector3f(1, 0, 0)
    uLINK[3].c2a        = Vector3f(0, 0, 0);
    uLINK[3].b          = Vector3f(Hip2RelHip1_X, Hip2RelHip1_Y, Hip2RelHip1_Z) +
                          uLINK[3].c2a - uLINK[2].c2a;
    uLINK[3].q          = 0;

    uLINK[4].name       = "LTHIGH";
    uLINK[4].sister     = 0;
    uLINK[4].child      = 5;
    uLINK[4].mother     = 3;
    uLINK[4].a          = Vector3f(1, 0, 0); // Vector3f(0, 1, 0)
    uLINK[4].c2a        = Vector3f(-ThighRelHip2_X, -ThighRelHip2_Y, -ThighRelHip2_Z);
    uLINK[4].b          = Vector3f(ThighRelHip2_X, ThighRelHip2_Y, ThighRelHip2_Z) +
                          uLINK[4].c2a - uLINK[3].c2a;
    uLINK[4].q = 0;

    uLINK[5].name       = "LSHANK";
    uLINK[5].sister     = 0;
    uLINK[5].child      = 6;
    uLINK[5].mother     = 4;
    uLINK[5].a          = Vector3f(1, 0, 0);
    uLINK[5].c2a        = Vector3f(0, -0.01, 0.045);
    uLINK[5].b          = Vector3f(ShankRelThigh_X, ShankRelThigh_Y, ShankRelThigh_Z) +
                          uLINK[5].c2a - uLINK[4].c2a;
    uLINK[5].q          = 0;

    uLINK[6].name       = "LANKLE";
    uLINK[6].sister     = 0;
    uLINK[6].child      = 7;
    uLINK[6].mother     = 5;
    uLINK[6].a          = Vector3f(1, 0, 0);
    uLINK[6].c2a        = Vector3f(0, 0, 0);
    uLINK[6].b          = Vector3f(AnkleRelShank_X, AnkleRelShank_Y, AnkleRelShank_Z) +
                          uLINK[6].c2a - uLINK[5].c2a;
    uLINK[6].q          = 0;

    uLINK[7].name       = "LFOOT";
    uLINK[7].sister     = 0;
    uLINK[7].child      = 0;
    uLINK[7].mother     = 6;
    uLINK[7].a          = Vector3f(0, 1, 0);
    uLINK[7].c2a        = Vector3f(-FootRelAnkle_X, -FootRelAnkle_Y, -FootRelAnkle_Z);
    uLINK[7].b          = Vector3f(FootRelAnkle_X, FootRelAnkle_Y, FootRelAnkle_Z) +
                          uLINK[7].c2a - uLINK[6].c2a;
    uLINK[7].q          = 0;

    uLINK[8].name       = "RHIP_1";
    uLINK[8].sister     = 0;
    uLINK[8].child      = 9;
    uLINK[8].mother     = 1;
    uLINK[8].a          = Vector3f(Hip1Axis_X, Hip1Axis_Y, Hip1Axis_Z);
    uLINK[8].c2a        = Vector3f(0, 0, 0);
    uLINK[8].b          = Vector3f(Hip1RelTorso_X, Hip1RelTorso_Y, Hip1RelTorso_Z) +
                          uLINK[8].c2a - uLINK[1].c2a;
    uLINK[8].q          = 0;

    uLINK[9].name       = "RHIP_2";
    uLINK[9].sister     = 0;
    uLINK[9].child      = 10;
    uLINK[9].mother     = 8;
    uLINK[9].a          = Vector3f(0, 1, 0); // Vector3f(1, 0, 0)
    uLINK[9].c2a        = Vector3f(0, 0, 0);
    uLINK[9].b          = Vector3f(Hip2RelHip1_X, Hip2RelHip1_Y, Hip2RelHip1_Z) +
                          uLINK[9].c2a - uLINK[8].c2a;
    uLINK[9].q          = 0;

    uLINK[10].name      = "RTHIGH";
    uLINK[10].sister    = 0;
    uLINK[10].child     = 11;
    uLINK[10].mother    = 9;
    uLINK[10].a         = Vector3f(1, 0, 0); // Vector3f(0, 1, 0)
    uLINK[10].c2a       = Vector3f(-ThighRelHip2_X, -ThighRelHip2_Y, -ThighRelHip2_Z);
    uLINK[10].b         = Vector3f(ThighRelHip2_X, ThighRelHip2_Y, ThighRelHip2_Z) +
                          uLINK[10].c2a - uLINK[9].c2a;
    uLINK[10].q         = 0;

    uLINK[11].name      = "RSHANK";
    uLINK[11].sister    = 0;
    uLINK[11].child     = 12;
    uLINK[11].mother    = 10;
    uLINK[11].a         = Vector3f(1, 0, 0);
    uLINK[11].c2a       = Vector3f(0, -0.01, 0.045);
    uLINK[11].b         = Vector3f(ShankRelThigh_X, ShankRelThigh_Y, ShankRelThigh_Z) +
                          uLINK[11].c2a - uLINK[10].c2a;
    uLINK[11].q         = 0;

    uLINK[12].name      = "RANKLE";
    uLINK[12].sister    = 0;
    uLINK[12].child     = 13;
    uLINK[12].mother    = 11;
    uLINK[12].a         = Vector3f(1, 0, 0);
    uLINK[12].c2a       = Vector3f(0, 0, 0);
    uLINK[12].b         = Vector3f(AnkleRelShank_X, AnkleRelShank_Y, AnkleRelShank_Z) +
                          uLINK[12].c2a - uLINK[11].c2a;
    uLINK[12].q         = 0;

    uLINK[13].name      = "RFOOT";
    uLINK[13].sister    = 0;
    uLINK[13].child     = 0;
    uLINK[13].mother    = 12;
    uLINK[13].a         = Vector3f(0, 1, 0);
    uLINK[13].c2a       = Vector3f(-FootRelAnkle_X, -FootRelAnkle_Y, -FootRelAnkle_Z);
    uLINK[13].b         = Vector3f(FootRelAnkle_X, FootRelAnkle_Y, FootRelAnkle_Z) +
                          uLINK[13].c2a - uLINK[12].c2a;
    uLINK[13].q         = 0;

    ForwardKinematics(1);

    NAO_A               = (uLINK[5].p - uLINK[4].p).Length();
    NAO_B               = (uLINK[6].p - uLINK[5].p).Length();	
    NAO_Dl              = uLINK[4].p - uLINK[1].p;
    NAO_Dr              = uLINK[10].p - uLINK[1].p;
    NAO_E               = uLINK[7].c2a;

    //uLINK[2].sister = 0;
}

void Kinematics::PrintLink(int j)
{
    if (j != 0)
    {
        printf("j = %d : %s, %.3f\n", j, uLINK[j].name.c_str(), gRadToDeg(uLINK[j].q));
        // cout << "j = " << j << " : " << uLINK[j].name << ", "
        //      << gRadToDeg(uLINK[j].q) << endl;
        PrintLink(uLINK[j].child);
        PrintLink(uLINK[j].sister);
    }
}

Kinematics::TIndex Kinematics::FindRoute(int to)
{
    Kinematics::TIndex idx;
    int i = uLINK[to].mother;

    if (i == 0)
    {
        // this should not happen
        assert(false);
    }
    else if (i == 1)
    {
        idx.push_back(to);
    }
    else
    {
        idx = FindRoute(i);
        idx.push_back(to);
    }

    return idx;
}

void Kinematics::ForwardKinematics(int j)
{
    if (j == 0) return ;
    if (j != 1)
    {
        int i = uLINK[j].mother;
        uLINK[j].p = uLINK[i].R.Rotate(uLINK[j].b) + uLINK[i].p;
        uLINK[j].R = uLINK[i].R * Rodrigues(uLINK[j].a, uLINK[j].q);
    }
    ForwardKinematics(uLINK[j].child);
    ForwardKinematics(uLINK[j].sister);
}

void Kinematics::ForwardVelocity(int j)
{
    if (j == 0) return ;
    if (j != 1)
    {
        int i = uLINK[j].mother;
        uLINK[j].v = uLINK[i].v + uLINK[i].w.Cross(uLINK[i].R.Rotate(uLINK[j].b));
        uLINK[j].w = uLINK[i].w + uLINK[i].R.Rotate(uLINK[j].a * uLINK[j].dq);
    }
    ForwardVelocity(uLINK[j].child);
    ForwardVelocity(uLINK[j].sister);
}

bool Kinematics::InverseKinematics(int to, const salt::Matrix& target)
{
    bool rec = true;
    const float lambda = 0.5f;
    TIndex idx = FindRoute(to);
    int jsize = idx.size();

    float err[6];
    float* dq = new float[jsize];

    // allot memory for Jacobian matrix
    float** J = new float*[6];
    for (int i = 0; i < 6; ++i)
    {
        J[i] = new float[jsize];
    }

    int n;
    for (n = 1; n <= 30; ++n)
    {
        CalcJacobian(J, idx);
        Matrix m = uLINK[to].R;
        m.Pos() = uLINK[to].p;
        CalcVWerr(err, target, m);

        float l = 0.0f;
        for (int i = 0; i < 6; ++i)
        {
            l += err[i] * err[i];
        }
        if (gSqrt(l) < 1e-3)
            break ;

        // solve J * dq = err
        // Solve(dq, J, err, 6, jsize);
        if (! Solve(dq, (const float **)J, (const float *)err, 6, jsize))
        {
            rec = false;
            goto over;
        }

#ifdef DEBUG
        printf("n = %d\n", n);
        printf("err: ");
        for (int i = 0; i < 6; ++i)
            printf("%.3f ", gRadToDeg(err[i]));
        printf("\n");
        printf("dq: ");
#endif
        for (int i = 0; i < jsize; ++i)
        {
            dq[i] *= lambda;
#ifdef DEBUG
            printf("%.3f ", gRadToDeg(dq[i]));
#endif
        }
#ifdef DEBUG
        printf("\n");
        printf("q: ");
#endif
        for (int nn = 0; nn < jsize; ++nn)
        {
            int j = idx.at(nn);
            uLINK[j].q += dq[nn];
            uLINK[j].q = gNormalizeRad(uLINK[j].q);
#ifdef DEBUG
            printf("%.3f ", gRadToDeg(uLINK[j].q));
#endif
        }
#ifdef DEBUG
        printf("\n");
        printf("J:\n");
        for (int i = 0; i < 6; ++i)
        {
            for (int j = 0; j < jsize; ++j)
                printf("%.3f ", J[i][j]);
            printf("\n");
        }
        printf("\n");
#endif
        ForwardKinematics(1);
    }
#ifdef DEBUG
    printf("n = %d\n", n);
#endif

over:
    for (int i = 0; i < 6; ++i)
    {
        delete [] J[i];
        J[i] = NULL;
    }
    delete [] J;
    delete [] dq;
    J = NULL;
    dq = NULL;

    return rec;
}

void Kinematics::CalcJacobian(float** J, const TIndex& idx)
{
    int jsize = idx.size();
    Vector3f target = uLINK[idx.at(jsize - 1)].p;

    for (int n = 0; n < jsize; ++n)
    {
        int j = idx.at(n);
        Vector3f a = uLINK[j].R * uLINK[j].a;
        Vector3f c = a.Cross(target - uLINK[j].p);
        J[0][n] = c.x();
        J[1][n] = c.y();
        J[2][n] = c.z();
        J[3][n] = a.x();
        J[4][n] = a.y();
        J[5][n] = a.z();
    }
}

void Kinematics::CalcVWerr(float err[6], const salt::Matrix& ref, const salt::Matrix& now)
{
    Vector3f perr = ref.Pos() - now.Pos();
    Matrix invNow = now;
    invNow.InvertRotationMatrix();
    // if (! InverseMatrix(invNow)) cout << "inverse matrix failed!" << endl;
    Matrix Rerr = invNow * ref;
    Vector3f werr = now.Rotate(RotToOmega(Rerr));
    err[0] = perr[0];
    err[1] = perr[1];
    err[2] = perr[2];
    err[3] = werr[0];
    err[4] = werr[1];
    err[5] = werr[2];
}

bool Kinematics::Solve(float* X, const float** A, const float* B, int row, int column)
{
    return svd(X, A, B, row, column);
}

bool Kinematics::IK_leg(const Matrix& hip,
                        float A, float B,
                        const Matrix& ankle,
                        float angle[6], bool left)
{
    Matrix invHip = hip;
    invHip.InvertRotationMatrix();
    Matrix invAnkle = ankle;
    invAnkle.InvertRotationMatrix();

    Vector3f r = invAnkle.Rotate(hip.Pos() - ankle.Pos());
    float C = r.Length();
    float cosKneePitch = (C * C - A * A - B * B) / (2.0f * A * B);
    float kneePitch = 0.0f;

    if (cosKneePitch >= 1.0f)
    {
        kneePitch = 0.0f;
        GetLog()->Error() << "cosKneePitch >= 1.0" << endl;
        return false;
    }
    else if (cosKneePitch <= -1.0f)
    {
        kneePitch = gPI;
        GetLog()->Error() << "cosKneePitch <= -1.0" << endl;
        return false;
    }
    else 
        kneePitch = gArcCos(cosKneePitch);

    kneePitch *= -1.0;

    float q6a = gArcSin((A / C) * gSin(gPI - kneePitch));
    float ankleRoll = -gArcTan2(r.x(), r.z());

    float anklePitch = gArcTan2(
                        r.y(),
                        gSign(r.z()) * gSqrt(r.x() * r.x() + r.z() * r.z())
                       ) - q6a;

    Matrix R = invHip * ankle;

    R.RotateY(-ankleRoll);
    R.RotateX(-anklePitch);
    R.RotateX(-kneePitch);

    float hipPitch = 0.0f;
    float hipRoll  = 0.0f;
    float hipYaw   = 0.0f;
    float cosHipYaw = 0.0f;
    float sinHipYaw = 0.0f;
    float cosHipRoll = 0.0f;
    float sinHipRoll = 0.0f;

    if (left)
    {
        hipPitch = gArcTan2(R(2,1) + R(0,1), R(2,2) + R(0,2));
        hipPitch = hipPitch >  gHalfPI ? hipPitch - gPI :
                   hipPitch < -gHalfPI ? hipPitch + gPI : hipPitch;

        float cosHipPitch   = gCos(hipPitch);
        float addCSHipRoll  = (R(2,2) + R(0,2)) / cosHipPitch;

        cosHipYaw = (R(2,0) - R(0,0)) / addCSHipRoll;
        sinHipYaw = R(1,0) / (-0.7071 * addCSHipRoll);
        hipYaw = gArcTan2(sinHipYaw, cosHipYaw);
        hipYaw = -hipYaw; // special in left leg
        hipYaw = hipYaw >  gHalfPI ? hipYaw - gPI :
                 hipYaw < -gHalfPI ? hipYaw + gPI : hipYaw;

        cosHipYaw = cos(hipYaw);

        float subCSHipRoll = 2 * R(0,0) - cosHipYaw * addCSHipRoll;
        cosHipRoll = addCSHipRoll + subCSHipRoll;
        sinHipRoll = addCSHipRoll - subCSHipRoll;
        hipRoll = gArcTan2(sinHipRoll, cosHipRoll);
        hipRoll = hipRoll >  gHalfPI ? hipRoll - gPI :
                  hipRoll < -gHalfPI ? hipRoll + gPI : hipRoll;
    }
    else
    {
        hipPitch = gArcTan2(R(2,1) + R(0,1), R(2,2) + R(0,2));
        hipPitch = hipPitch >  gHalfPI ? hipPitch - gPI :
                   hipPitch < -gHalfPI ? hipPitch + gPI : hipPitch;

        float cosHipPitch   = gCos(hipPitch);
        float subCSHipRoll  = (R(2,2) - R(0,2)) / cosHipPitch;

        cosHipYaw = (R(2,0) + R(0,0)) / subCSHipRoll;
        sinHipYaw = R(1,0) / (0.7071 * subCSHipRoll);
        hipYaw = gArcTan2(sinHipYaw, cosHipYaw);
        hipYaw = hipYaw >  gHalfPI ? hipYaw - gPI :
                 hipYaw < -gHalfPI ? hipYaw + gPI : hipYaw;
        cosHipYaw = cos(hipYaw);

        float addCSHipRoll = 2 * R(0,0) - cosHipYaw * subCSHipRoll;
        cosHipRoll = addCSHipRoll + subCSHipRoll;
        sinHipRoll = addCSHipRoll - subCSHipRoll;
        hipRoll = gArcTan2(sinHipRoll, cosHipRoll);
        hipRoll = hipRoll >  gHalfPI ? hipRoll - gPI :
                  hipRoll < -gHalfPI ? hipRoll + gPI : hipRoll;
    }

    angle[0] = hipYaw;
    angle[1] = hipRoll;
    angle[2] = hipPitch;
    angle[3] = kneePitch;
    angle[4] = anklePitch;
    angle[5] = ankleRoll;

    return true;
}

bool Kinematics::IK_leg(const salt::Matrix& torso, const salt::Matrix& foot,
                        float angle[6], bool left)
{
    uLINK[1].R = torso;
    uLINK[1].p = torso.Pos();

    Vector3f D;
    int low, high;
    if (left)
    {
        D = NAO_Dl;
        low = 2;
        high = 7;
    }
    else
    {
        D = NAO_Dr;
        low = 8;
        high = 13;
    }

    Matrix hip = torso;
    hip.Pos() = torso * D;
    Matrix ankle = foot;
    ankle.Pos() = foot * NAO_E;

    if (! IK_leg(hip, NAO_A, NAO_B, ankle, angle, left)) return false;
    for (int i = 0; i < 6; ++i)
    {
        uLINK[i + low].q = angle[i];
    }

    ForwardKinematics(1); // apply the joint angles before InverseKinematics
    if (! InverseKinematics(high, ankle)) return false;
    for (int i = 0; i < 6; ++i) angle[i] = uLINK[i + low].q;
    return true;
}

void Kinematics::CalcJointVel(const salt::Vector3f& vb, const salt::Vector3f& wb,
                              const salt::Vector3f& vt, const salt::Vector3f& wt,
                              float vel[7], bool left)
{
    int to = left ? 7 : 13;
    TIndex idx = FindRoute(to);
    int jsize = idx.size();

    Vector3f vd, vw;
    vd = vt - vb - wb.Cross(uLINK[to].p - uLINK[1].p);
    vw = wt - wb;

    float v[7];
    v[0] = vd.x(); v[1] = vd.y(); v[2] = vd.z();
    v[3] = vw.x(); v[4] = vw.y(); v[5] = vw.z();
    v[6] = 0.0;

    // allot memory for Jacobian matrix
    float** J = new float*[6];
    for (int i = 0; i < 6; ++i)
    {
        J[i] = new float[jsize];
    }

    CalcJacobian(J, idx);
    Solve(vel, (const float **)J, (const float *)v, 6, jsize);

    for (int i = 0; i < 6; ++i)
    {
        delete [] J[i];
        J[i] = NULL;
    }
    delete [] J;
    J = NULL;
}

