#ifndef KINEMATICS_H
#define KINEMATICS_H

#include <string>
#include <vector>
#include <zeitgeist/class.h>
#include <salt/vector.h>
#include <salt/matrix.h>

class Kinematics : public zeitgeist::Leaf
{
public:
    typedef std::vector<int> TIndex;

    struct Link {
        std::string name;
        int id;
        int sister;
        int child;
        int mother;
        salt::Vector3f a;     // axis
        salt::Vector3f c2a;   // geometry center to anchor
        salt::Vector3f b;     // relative position to mother's anchor
        float q;              // joint angle
        float dq;
        salt::Vector3f p;
        salt::Matrix   R;
        salt::Vector3f v;
        salt::Vector3f w;
    };

    Link uLINK[14];

public:
    Kinematics();

    ~Kinematics();

    void PrintLink(int j);

    bool IK_leg(const salt::Matrix& torso, const salt::Matrix& foot,
                float angle[6], bool left);

    void CalcJointVel(const salt::Vector3f& vb, const salt::Vector3f& wb,
                      const salt::Vector3f& vt, const salt::Vector3f& wt,
                      float vel[6], bool left);

private:
    void SetupLinks();

    /** A*x = B, size: A--row*column, x--colunm*1, B--row*1 */
    static bool Solve(float* X, const float** A, const float* B, int row, int column);

    static void CalcVWerr(float err[6], const salt::Matrix& ref, const salt::Matrix& now);

    TIndex FindRoute(int to);

    void ForwardKinematics(int j);

    bool InverseKinematics(int to, const salt::Matrix& target);

    void CalcJacobian(float** J, const TIndex& idx);

    void ForwardVelocity(int j);

    bool IK_leg(const salt::Matrix& hip,
                float A, float B,
                const salt::Matrix& ankle,
                float angle[6], bool left);

private:
    /** foots */
    float FootLength;
    float FootWidth;
    float FootHeight;
    float FootMass;

    float FootRelAnkle_X;
    float FootRelAnkle_Y;
    float FootRelAnkle_Z;

    /** ankles */
    float AnkleRelShank_X;
    float AnkleRelShank_Y;
    float AnkleRelShank_Z;


    /** shanks */
    float ShankLength;
    float ShankWidth;
    float ShankHeight;
    float ShankMass;

    float ShankRelThigh_X;
    float ShankRelThigh_Y;
    float ShankRelThigh_Z;

    /** thighs */
    float ThighLength;
    float ThighWidth;
    float ThighHeight;
    float ThighMass;

    float ThighRelHip2_X;
    float ThighRelHip2_Y;  //set 0.005 to check the tri-angle algorithm
    float ThighRelHip2_Z;

    /** hips */
    float Hip1Axis_X ;
    float Hip1Axis_Y ;
    float Hip1Axis_Z ;

    float Hip1RelTorso_X;
    float Hip1RelTorso_Y;
    float Hip1RelTorso_Z;

    float Hip2RelHip1_X;
    float Hip2RelHip1_Y;
    float Hip2RelHip1_Z;

    /** torso */
    float TorsoLength;
    float TorsoWidth;
    float TorsoHeight;
    float TorsoMass;

    float TorsoInitX;
    float TorsoInitY;
    float TorsoInitZ;

    /**  distance and vector */
    float NAO_A;
    float NAO_B;
    salt::Vector3f NAO_Dl;
    salt::Vector3f NAO_Dr;
    salt::Vector3f NAO_E;
};

DECLARE_CLASS(Kinematics);

#endif

