#include "network.h"
#include <errno.h>

using namespace rcss::net;
using namespace std;

Network::Network()
{
    mHost = "127.0.0.1";
    mPort = 3100;
}

Network::~Network()
{
}

void Network::OnLink()
{
}

void Network::OnUnLink()
{
}

bool Network::Init()
{
    cout << "connecting to TCP " << mHost << ":" << mPort << "\n";

    try
    {
        Addr local(INADDR_ANY,INADDR_ANY);
        mSocket.bind(local);
    }

    catch (BindErr error)
    {
        cerr << "failed to bind socket with '"
            << error.what() << "'" << endl;

        mSocket.close();
        return false;
    }

    try
    {
        Addr server(mPort,mHost);
        mSocket.connect(server);
    }

    catch (ConnectErr error)
    {
        cerr << "connection failed with: '"
            << error.what() << "'" << endl;
        mSocket.close();
        return false;
    }

    return true;
}

void Network::Done()
{
    mSocket.close();
    cout << "closed connection to " << mHost << ":" << mPort << "\n";
}

bool Network::SelectInput()
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(mSocket.getFD(), &readfds);

#ifdef WIN32
    int maxFd = 0;
#else
    int maxFd = mSocket.getFD()+1;
#endif
    return select(maxFd,&readfds, 0, 0, 0) > 0;
}

bool Network::PutMessage(const string& msg)
{
    if (msg.empty())
    {
        return true;
    }

//     // prefix the message with it's payload length
  //   unsigned int len = static_cast<unsigned int>(htonl(msg.size()));
  //  mSocket.send((const char*)&len, sizeof(unsigned int));
  //  int rval = mSocket.send(msg.data(), msg.size());

    unsigned int len = htonl(msg.size());
  string prefix((const char*)&len,sizeof(unsigned int));
  string str = prefix + msg;
  int rval = mSocket.send(str.data(),str.size());

    if (rval < 0)
    {
        cerr
            << "(PutMessage) ERROR: "
            << " send returned error '"
            << strerror(errno) << "' " << endl;
        return false;
    }
    return true;
}


bool Network::GetMessage(string& msg)
{
    static char buffer[16 * 1024];

    unsigned int bytesRead = 0;
    while(bytesRead < sizeof(unsigned int))
    {
        SelectInput();

        int readResult = mSocket.recv(buffer + bytesRead, sizeof(unsigned int) - bytesRead);
        if(readResult <= 0)
        {
            cerr 
                << "(GetMessage) ERROR: "
                << " read returned error '"
                << strerror(errno) << "' " << endl;
            // continue;
            return false;
        }
        bytesRead += readResult;
    }

    //cerr << "buffer = |" << string(buffer+1) << "|\n";
    //cerr << "bytesRead = |" << bytesRead << "|\n";
    //cerr << "Size of buffer = |" << sizeof(buffer) << "|\n";
    //cerr << "buffer = |" << buffer << "|\n";
    //cerr << "buffer[5] = |" << buffer[5] << "|\n";
    //printf ("xxx-%s\n", buffer+5);

    // msg is prefixed with it's total length
    unsigned int msgLen = ntohl(*(unsigned int*)buffer);
    // cerr << "GM 6 / " << msgLen << " (bytesRead " << bytesRead << ")\n";
    if(sizeof(unsigned int) + msgLen > sizeof(buffer)) {
        cerr << "too long message; aborting" << endl;
        abort();
    }

    // read remaining message segments
    unsigned int msgRead = bytesRead - sizeof(unsigned int);

    //cerr << "msgRead = |" << msgRead << "|\n";

    char *offset = buffer + bytesRead;

    while (msgRead < msgLen)
    {
        if (! SelectInput())
        {
            return false;
        }

        int readLen = sizeof(buffer) - msgRead;
        if(readLen > msgLen - msgRead)
            readLen = msgLen - msgRead;

        int readResult = mSocket.recv(offset, readLen);
        if (readResult <= 0)
        {
            cerr 
                << "(GetMessage) ERROR: "
                << " read returned error '"
                << strerror(errno) << "' " << endl;
    
            // continue;
            return false;
        }
        msgRead += readResult;
        offset += readResult;
        //cerr << "msgRead = |" << msgRead << "|\n";
    }

    // zero terminate received data
    (*offset) = 0;

    msg = string(buffer+sizeof(unsigned int));

    // DEBUG
    // cout << msg << endl;

    return true;
}
